;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :cm)

;;;
;;; Any port of this midi code needs to implement a low level 
;;; interface of the following functions:
;;;
;;;
;;; %midi-open (port)               open midi
;;; %midi-close ()                  close midi
;;; %midi-stop-time ()              stop the time
;;; %midi-start-time ()             start the time
;;; %midi-get-time ()               get millisecond time
;;; %midi-set-time (milli)          set millisecond time
;;; %midi-read-messages ()          read pending input
;;; %midi-write-message (msg qtime) write midi message
;;; %midi-flush-transmit ()         clear pending output queue
;;; %midi-flush-receive  ()         clear pending input queue
;;; %midi-hush  ()                  clear and stop sound
;;; %midi-all-notes-off (time)      schedule note offs at time
;;; %midi-default-channel-map (port)
;;;

#-(or mcl (and excl sgi) (and excl next) (and akcl |NeXT|) (and clisp next))
(progn
  (warn "~%MIDI real time not implemented for this machine/os/lisp.")
  (unless (= internal-time-units-per-second 1000)
    (warn "~%INTERNAL-TIME-UNITS-PER-SECOND is ~A in this Lisp.  Fix ~
           midi-get-time and midi-start-time to return MILLISECONDS!"
          internal-time-units-per-second))
  (defvar .base-time. 0)
  (defun %midi-open (port) 
    (setf *midi-open* port)
    +midi-success+)
  (defun %midi-get-chan-map () nil)
  (defun %midi-set-chan-map (map) map)
  (defun %midi-close () +midi-success+)
  (defun %midi-start-time () 
    (%midi-set-time 0)
    +midi-success+)
  (defun %midi-stop-time () 
    +midi-success+)
  (defun %midi-get-time () 
    (- (get-internal-real-time) .base-time.))
  (defun %midi-set-time (int)
    (setf .base-time. (+ (get-internal-real-time) int))
    +midi-success+)
  (defun %midi-read-messages () +midi-success+)
  (defun %midi-write-message (int1 int2) int1 int2 +midi-success+)
  (defun %midi-flush-transmit () +midi-success+)
  (defun %midi-flush-receive () +midi-success+)
  (defun %midi-hush () +midi-success+)
  (defun %midi-all-notes-off (int) int +midi-success+)
  (defun %midi-default-channel-map (port) port (default-midi-channel-map))
  )

;;;
;;; opening/closing
;;;

(defun midi-open (&key port chanmap)
  (declare (ignore chanmap))
  (when port
    (unless (midi-port-reference? port)
      #-tl (error "~S is not a MIDI port." port)
      #+tl (setf port
                 (ask-port :prompt "Open MIDI on port: "
                           :type 'midi :check t
                           :input (format nil "~A" port)))
      (when (eq port ':aborted)
        (error "Can't process MIDI: No port specifed."))
      (unless (or (not (midi-open?))
                  (eql (midi-port-number *midi-open*)
                       (midi-port-number port)))
        (midi-close))))
  (or *midi-open*
      (with-%midi-success-checked 
        (%midi-open (if port (midi-port-number port) -1))
        (setf *default-midi-channel-map*
              (%midi-default-channel-map *midi-open*))
        *midi-open*)))
  
(defun midi-close (&optional port)
  (declare (ignore port))
  (when *midi-open*
    (with-%midi-success-checked (%midi-close)
      (setf *midi-open* nil)
      (setf *default-midi-channel-map* (default-midi-channel-map ))
      
      t)))

(defun midi-open? () *midi-open*)

#+tl
(defun ask-port (&key prompt (input "") default (abort-ok t) (check t)
                      (type 'midi) (stream t) (abort-value :aborted))
  (let (port)
    (if (and check (setf port (midi-open?)))
        port
      (progn
        (unless prompt
          (setf prompt (format nil "~:(~A~) port to open: " type)))
        (tl:ask-user :prompt prompt :input input :default default
                     :stream stream :null-ok (and default ':ask) 
                     :predicate #'midi-port-reference?
                     :abort-ok abort-ok :abort-value abort-value)))))
;;;
;;; timer functions
;;;

(defun midi-stop-timer ()
  (%midi-stop-time))

(defun midi-start-timer ()
  (%midi-start-time))
		
(defun midi-get-time ()
  (let ((val (%midi-get-time)))
    (when (< val +midi-success+)
      (warn-user "~&%midi-get-time failed with error: ~s" val))
    val))

(defun midi-set-time (time)
  (with-%midi-success-checked (%midi-set-time time)
    time))

;;;
;;; resetting
;;;

(defun midi-hush (&optional (flush t))
  (declare (ignore flush))
  (with-%midi-success-checked (%midi-hush)
    t))

(defun midi-all-notes-off (&optional (time 0))
  (with-%midi-success-checked (%midi-all-notes-off time)
    t))
 
;;;
;;; writing
;;;
						 
(defun midi-write-message (message &optional (time 0) message-data)
  (let ((flg (%midi-write-message message time)))
    (declare (fixnum flg))
    (when message-data
      (loop for m in message-data
	    while (= flg +midi-success+)
	    do
	 (setf flg (%midi-write-message m time))))
    (if (= flg +midi-success+)
	t
      (progn (warn "%midi-write-message failed with error: ~s" flg)
	     nil))))  

;;;
;;;  reading
;;;						 

(defun midi-set-read-hook (fn)
  (setf *midi-read-hook* fn))

(defun midi-read-hook ()
  *midi-read-hook*)
						 
(defun midi-read-messages (&optional (fn *midi-read-hook*))
  (declare (optimize (speed 3) (safety 0)))
  (let ((*midi-read-hook* fn))
    (with-%midi-success-checked (%midi-read-messages) 
      t)))

(defvar *midi-receive-process* nil)
(defvar *midi-receiving* nil)

(defun midi-receive (receiver 
                     &rest args
                     &key (async? #+(or processes multiprocessing) t)
                     &allow-other-keys)
  (if (midi-open?)
    (if *midi-receiving*
      (warn-user "~%MIDI: receiving already enabled.")
      (if async?
        (setf *midi-receive-process*
              (apply #'process-run-function 
                     (list :name "MIDI Listener"
                           :priority (if (eq async? t) 0 async?))
                     #'midi-receive-aux 
                     receiver args))
        (apply #'midi-receive-aux receiver args)))
    (warn-user "~%MIDI: no port open.")))

(defun midi-receive-aux (receiver &key keynum duration stop-if 
                                  stop-if-not initializer finalizer
                                  &allow-other-keys)
  (declare (optimize (speed 3) (safety 0)))
  (let ((milli? (eq (clock-mode) ':milliseconds))
        hook base end)
    (macrolet ((checkdup (&rest args)
                 `(when (or ,@args)
                    (error ":KEYNUM, :DURATION, :STOP-IF, :STOP-IF-NOT ~
                            are mutually exclusive keywords."))))
      (cond (duration
             (checkdup keynum stop-if stop-if-not)
             (unless (and milli? (integerp duration)) 
               (setf duration (milliseconds duration)))
             (setf hook
                   #'(lambda (msg time)
                       (declare (fixnum msg time)
                                (optimize (speed 3) (safety 0)))
                       (unless base 
                         (setf base time end (+ time duration)))
                       (if (> time (the fixnum end))
                         (midi-stop-receiving)
                         (funcall (the function receiver)
                                  msg (- time base))))))
            (keynum
             (checkdup duration stop-if-not stop-if)
             (setf hook 
                   #'(lambda (msg time)
                       (declare (fixnum msg time)
                                (optimize (speed 3) (safety 0)))
                       (if (= (the fixnum (note-on-key msg))
                              (the fixnum keynum))
                         (if (or (= 0 (the fixnum (note-on-velocity msg))) 
                                 (note-off-p msg))
                           (funcall (the function receiver) 
                                    msg 
                                    (if base (- time base)
                                        (progn (setf base time) 0)))
                           (midi-stop-receiving))))))
            (stop-if
             (checkdup keynum duration stop-if-not)
             (setf hook
                   #'(lambda (msg time) 
                       (declare (fixnum msg time) 
                                (optimize (speed 3) (safety 0)))
                       (if (funcall (the function stop-if)
                                    msg 
                                    (if base (- time base)
                                        (progn (setf base time) 0)))
                         (midi-stop-receiving)
                         (funcall (the function receiver)
                                  msg 
                                  (if base (- time base)
                                      (progn (setf base time) 0)))))))
            (stop-if-not
             (checkdup keynum duration stop-if)
             (setf hook
                   #'(lambda (msg time) 
                       (declare (fixnum msg time)
                                (optimize (speed 3) (safety 0)))
                       (if (funcall (the function stop-if-not)
                                    msg 
                                    (if base (- time base)
                                        (progn (setf base time) 0)))
                         (funcall (the function receiver)
                                  msg
                                  (if base (- time base)
                                      (progn (setf base time) 0)))
                         (midi-stop-receiving)))))
            (t
             (setf hook receiver))))
    (unwind-protect
      (let ((*midi-read-hook* hook))
       ; (%midi-flush-receive)
        (when initializer (funcall initializer))
        (setf *midi-receiving* t)
        (loop while *midi-receiving*
              for flag = (%midi-read-messages)
              do
              (unless (= (the fixnum flag) +midi-success+)
                (warn "%midi-receive failed with error: ~D" flag)
                (setf *midi-receiving* nil))))
      (setf *midi-receiving* nil))
    ;(%midi-flush-receive)
    (when finalizer (funcall finalizer))))

(defun midi-stop-receiving ()
  (setf *midi-receiving* nil))

(defun make-recording-closure (&optional thread)
  (unless thread
    (setf thread (make-object 'thread))
    (update-top-level ':add thread))
  (let ((base nil)
        (bag (slot-value thread 'elements)))
    (if (eq (clock-mode) ':milliseconds)
      #'(lambda (mm mt)
          (%midi-write-message mm mt)
          (append-bag (make-instance 'midi-message 
                        :message mm :rhythm 0
                        :start (if base (- mt base)
                                   (progn (setf base mt) 0)))
                      bag))
      #'(lambda (mm mt)
          (%midi-write-message mm mt)
          (append-bag (make-instance 'midi-message 
                        :message mm :rhythm 0
                        :start (if base (seconds (- mt base))
                                   (progn (setf base mt) 0.0)))
                      bag)))))


